/*
 *	GENERA.C
 *	LaMoSca (LAboratorio di MOtori per SCAcchi)
 *	versione 0.07
 *	2001 Pietro Valocchi
 */


#include <stdio.h>
#include <string.h>
#include "Common.h"
#include "Extern.h"

/* memoria mosse eseguite */
Casa History[MAXLIVELLI+1];


/* posizione iniziale della scacchiera */
int ColoriIniziali[NUMEROCASE] = {
	NERO,   NERO,   NERO,   NERO,   NERO,   NERO,   NERO,   NERO,
	NERO,   NERO,   NERO,   NERO,   NERO,   NERO,   NERO,   NERO,
	VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,
	VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,
	VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,
	VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,  VUOTO,
	BIANCO, BIANCO, BIANCO, BIANCO, BIANCO, BIANCO, BIANCO, BIANCO,
	BIANCO, BIANCO, BIANCO, BIANCO, BIANCO, BIANCO, BIANCO, BIANCO
};

int PezziIniziali[NUMEROCASE] = {
	TORRE,  CAVALLO, ALFIERE, REGINA, RE,     ALFIERE, CAVALLO, TORRE,
	PEDONE, PEDONE,  PEDONE,  PEDONE, PEDONE, PEDONE,  PEDONE,  PEDONE,
	VUOTO,  VUOTO,   VUOTO,   VUOTO,  VUOTO,  VUOTO,   VUOTO,   VUOTO,
	VUOTO,  VUOTO,   VUOTO,   VUOTO,  VUOTO,  VUOTO,   VUOTO,   VUOTO,
	VUOTO,  VUOTO,   VUOTO,   VUOTO,  VUOTO,  VUOTO,   VUOTO,   VUOTO,
	VUOTO,  VUOTO,   VUOTO,   VUOTO,  VUOTO,  VUOTO,   VUOTO,   VUOTO,
	PEDONE, PEDONE,  PEDONE,  PEDONE, PEDONE, PEDONE,  PEDONE,  PEDONE,
	TORRE,  CAVALLO, ALFIERE, REGINA, RE,     ALFIERE, CAVALLO, TORRE
};


/* Indica se un pezzo puo' muovere di piu' case alla volta
   0 PEDONE, 1 CAVALLO, 2 ALFIERE, 3 TORRE, 4 REGINA, 5 RE 
   Il pedone e' trattato separatamente */
BOOL Trascina[NUMEROPEZZI] = {
	FALSE, FALSE, TRUE, TRUE, TRUE, FALSE
};

/* Numero delle direzioni in cui un pezzo puo' muovere
   0 PEDONE, 1 CAVALLO, 2 ALFIERE, 3 TORRE, 4 REGINA, 5 RE 
   Il pedone e' trattato separatamente */
int NumeroDirezioni[NUMEROPEZZI] = {
	0, 8, 4, 4, 8, 8
};

/* Direzioni dei vari pezzi in una matrice 10*12 (vedi matrice Da12A8[])
   0 PEDONE, 1 CAVALLO, 2 ALFIERE, 3 TORRE, 4 REGINA, 5 RE
   Es.: cavallo -21 vuol dire 2 righe sotto ed una colonna a sinistra
   Il pedone e' trattato separatamente */
int OffsetPezzi[NUMEROPEZZI][CASEPERRIGA] = {
	{   0,   0,   0,  0, 0,  0,  0,  0 },
	{ -21, -19, -12, -8, 8, 12, 19, 21 },
	{ -11,  -9,   9, 11, 0,  0,  0,  0 },
	{ -10,  -1,   1, 10, 0,  0,  0,  0 },
	{ -11, -10,  -9, -1, 1,  9, 10, 11 },
	{ -11, -10,  -9, -1, 1,  9, 10, 11 }
};

/* Da12A8[] e' una matrice 10*12 utilizzata per verificare la legalita' 
   delle mosse. E' ottenuta aggiungendo ad una matrice 8*8 una 
   colonna per lato e due righe superiori ed inferiori con valori pari a -1,
   che semplificano il test di mosse che cadono al di fuori della scacchiera.
   Sono necessarie due righe per tener conto delle due case della mossa del cavallo,
   cosa che orizzontalmente e' fornita dell'unione delle colonne destra e sinistra.
   Le celle diverse da -1 rappresentano gli indici nella scacchiera 8*8 */

int Da12A8[120] = {
	 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	 -1,  0,  1,  2,  3,  4,  5,  6,  7, -1,
	 -1,  8,  9, 10, 11, 12, 13, 14, 15, -1,
	 -1, 16, 17, 18, 19, 20, 21, 22, 23, -1,
	 -1, 24, 25, 26, 27, 28, 29, 30, 31, -1,
	 -1, 32, 33, 34, 35, 36, 37, 38, 39, -1,
	 -1, 40, 41, 42, 43, 44, 45, 46, 47, -1,
	 -1, 48, 49, 50, 51, 52, 53, 54, 55, -1,
	 -1, 56, 57, 58, 59, 60, 61, 62, 63, -1,
	 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
	 -1, -1, -1, -1, -1, -1, -1, -1, -1, -1
};

/* Per ogni casa della scacchiera 8*8, la matrice Da8A12[] fornisce 
   l'indice della corrispondente casa nella matrice 10*12 Da12A8[].
   L'utilizzo sara' del tipo (vedi la funzione GeneraPseudo() )
   CasaDestinazione = 
   Da12A8[Da8A12[CasaDestinazione] + OffsetPezzi[Pezzi[CasaCorrente]][Direzione]];
*/

int Da8A12[NUMEROCASE] = {
	21, 22, 23, 24, 25, 26, 27, 28,
	31, 32, 33, 34, 35, 36, 37, 38,
	41, 42, 43, 44, 45, 46, 47, 48,
	51, 52, 53, 54, 55, 56, 57, 58,
	61, 62, 63, 64, 65, 66, 67, 68,
	71, 72, 73, 74, 75, 76, 77, 78,
	81, 82, 83, 84, 85, 86, 87, 88,
	91, 92, 93, 94, 95, 96, 97, 98
};


/* SalvaMossa() inserisce una mossa nello stack */
void SalvaMossa(int Da, int A) {

	StackMosse[InizioMosse[Livello]].Da = (char)Da;
	StackMosse[InizioMosse[Livello]].A  = (char)A;
	InizioMosse[Livello]++;
}


/* GeneraPseudo() genera mosse pseudolegali per la posizione corrente */
void GeneraPseudo() {
	int i, CasaDestinazione, Direzione;

	/* azzera il contatore delle mosse */
	InizioMosse[Livello] = InizioMosse[Livello-1];

	/* cerca tutte le case per pezzi del colore che deve muovere */
	for (i = 0; i < NUMEROCASE; i++) {
		if (B[i].Colore == ChiMuove) {
			if (B[i].Pezzo == PEDONE) {
				if (ChiMuove == BIANCO) {
					/* mosse per il pedone bianco */
					if (COLONNA(i) != 0 && B[i - 9].Colore == NERO)
						/* mangia in avanti a sinistra */
						SalvaMossa(i, i - 9);
					if (COLONNA(i) != 7 && B[i - 7].Colore == NERO)
						/* mangia in avanti a destra */
						SalvaMossa(i, i - 7);
					if (B[i - CASEPERRIGA].Colore == VUOTO) {
						/* avanti una casa */
						SalvaMossa(i, i - 8);
						if (i >= 48 && B[i - 16].Colore == VUOTO)
							/* avanti due case */
							SalvaMossa(i, i - 16);
					}
				}
				else {
					/* mosse per il pedone nero */
					if (COLONNA(i) != 0 && B[i + 7].Colore == BIANCO)
						SalvaMossa(i, i + 7);
					if (COLONNA(i) != 7 && B[i + 9].Colore == BIANCO)
						SalvaMossa(i, i + 9);
					if (B[i + CASEPERRIGA].Colore == VUOTO) {
						SalvaMossa(i, i + 8);
						if (i <= 15 && B[i + 16].Colore == VUOTO)
							SalvaMossa(i, i + 16);
					}
				}
			} else {
				/* per i pezzi differenti dal pedone */
				/* genera mosse per ciascuna delle direzioni legali */
				for (Direzione = 0; Direzione < NumeroDirezioni[B[i].Pezzo]; Direzione++) {
					/* casa destinazione inizializzata con la corrente */
					CasaDestinazione = i;
					while (TRUE) {
						/* calcola la casa destinazione sommando l'offset
						per il pezzo corrente alla casa corrente.
						Si passa per la matrice 10*12 per verificare che
						la mossa non cada fuori dalla scacchiera */
						CasaDestinazione = Da12A8[Da8A12[CasaDestinazione] + 
							OffsetPezzi[B[i].Pezzo][Direzione]];
						/* mossa fuori dalla scacchiera */
						if (CasaDestinazione == -1)
							break;
						/* mossa su una casa con un pezzo avversario */
						if (B[CasaDestinazione].Colore != VUOTO) {
							if (B[CasaDestinazione].Colore != ChiMuove)
								SalvaMossa(i, CasaDestinazione);
							break;
						}
						/* mossa su una casa vuota */
						SalvaMossa(i, CasaDestinazione);
						/* se il pezzo puo' muovere di piu' case, continua */
						if (!Trascina[B[i].Pezzo])
							break;
					}
				}
			}
		}
	}
}


/* Inizializza() inizializza la scacchiera */
void Inizializza() {
	int i;

	for (i = 0; i < NUMEROCASE; i++) {
		B[i].Colore = ColoriIniziali[i];
		B[i].Pezzo  = PezziIniziali[i];
	}
	ChiMuove = BIANCO;
	ColoreProgramma = VUOTO;
	LivelloMassimo = MAXLIVELLI;
}


/* Muovi() riporta una mossa sulla scacchiera */
void Muovi(Mossa *M) {

	/* salva informazioni sulla casa destinazione 
	   per tornare eventualmente indietro */
	History[Livello-1].Colore = B[M->A].Colore;
	History[Livello-1].Pezzo  = B[M->A].Pezzo;

	/* esegui la mossa */
	B[M->A].Colore  = B[M->Da].Colore;
	B[M->A].Pezzo   = B[M->Da].Pezzo;
	B[M->Da].Colore = VUOTO;
	B[M->Da].Pezzo  = VUOTO;

	/* cambia colore */
	ChiMuove ^= 1;

	/* incrementa il contatore dei livelli */
	Livello++;

}


/* TornaIndietro() ripristina una mossa salvata */
void TornaIndietro(Mossa *M) {

	/* decrementa il contatore dei livelli */
	Livello--;

	/* cambia colore */
	ChiMuove ^= 1;

	/* ripristina la vecchia mossa */
	B[M->Da].Colore = B[M->A].Colore;
	B[M->Da].Pezzo  = B[M->A].Pezzo;
	B[M->A].Colore  = History[Livello-1].Colore;
	B[M->A].Pezzo   = History[Livello-1].Pezzo;

}


/* MossaAvversario() esegue una mossa dell'avversario */
BOOL MossaAvversario(char * Input) {
	BOOL Trovato = FALSE;
	int i, Da, A, ch;
	Mossa M;

	/* calcola dai primi due caratteri l'indice della casa di partenza */
	Da = (Input[0] - 'a') + CASEPERRIGA * (CASEPERRIGA - (Input[1] - '0'));
	/* calcola dai caratteri rimanenti l'indice della casa di destinazione */
	A  = (Input[2] - 'a') + CASEPERRIGA * (CASEPERRIGA - (Input[3] - '0'));

	/* genera lo stack delle mosse per l'avversario salvando il numero mosse generate */
	Livello = 1;
	GeneraPseudo();

	/* controlla che la mossa sia legale cercandola nello stack */
	for (i = InizioMosse[Livello-1]; i < InizioMosse[Livello]; i++) {
		if ((StackMosse[i].Da == Da) && (StackMosse[i].A == A)) {
			Muovi(&StackMosse[i]);
			Trovato = TRUE;

			/* verifica se si tratta di una promozione */
			if ((B[A].Pezzo == PEDONE) && ((RIGA(A) == 0) || (RIGA(A) == 7))) {
				ch = Input[4];
				B[StackMosse[i].A].Pezzo = strchr(SimboliPezziNero, ch) - SimboliPezziNero;
			}

			/* se  la prima mossa dell'avversario, fa partire il programma */
			if ((ForceMode == FALSE) && (ColoreProgramma == VUOTO))
				ColoreProgramma = ChiMuove;
			break;

		}
	}

	/* controlla se si tratta di un arrocco */
	if (ChiMuove == BIANCO) {
		if (!strcmp(Input, "e1g1")) {
			M.Da = H1; M.A = F1;
			Muovi(&M);
			M.Da = E1; M.A = G1;
			Muovi(&M);
			Trovato = TRUE;
		}
		if (!strcmp(Input, "e1c1")) {
			M.Da = A1; M.A = D1;
			Muovi(&M);
			M.Da = E1; M.A = C1;
			Muovi(&M);
			Trovato = TRUE;
		}
	} else {
		if (!strcmp(Input, "e8g8")) {
			M.Da = H8; M.A = F8;
			Muovi(&M);
			M.Da = E8; M.A = G8;
			Muovi(&M);
			Trovato = TRUE;
		}
		if (!strcmp(Input, "e8c8")) {
			M.Da = A8; M.A = D8;
			Muovi(&M);
			M.Da = E8; M.A = C8;
			Muovi(&M);
			Trovato = TRUE;
		}
	}

	return Trovato;
}


/* MossaProgramma() genera una mossa per il computer */
BOOL MossaProgramma(char *CoordinateMossa) {

	Mossa M;
	int ValorePosizione;
	BOOL Trovata;

	Livello = 1;

	/* massimizza il punteggio del programma */
	Trovata = NegaMax(&M, &ValorePosizione);

	if (Trovata) {
		/* esegui la mossa */
		Muovi(&M);

		/* crea la stringa con le coordinate della mossa corrente */
		sprintf(CoordinateMossa, "%c%d%c%d",
			COLONNA(M.Da)+ 'a', 8 - RIGA(M.Da), 
			COLONNA(M.A) + 'a', 8 - RIGA(M.A));
	}

	return Trovata;
}


